/*----------------------------------------------------------------------------/
/  SP850ES - Flash Programmer for V850ES microcontrollers  R0.04a
/  http://elm-chan.org/
/-----------------------------------------------------------------------------/
/ SP850ES is a flash programming software for Renesas V850ES family MCUs.
/ It is a Free Software opened under license policy of GNU GPL.

   Copyright (C) 2015, ChaN, all right reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.

   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.

   * Neither the name of the copyright holders nor the names of
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. 
*/


#include <stdio.h>
#include <string.h>
#include <windows.h>


#define	INIFILE		"sp850es.ini"
#define MAX_FLASH	0x100000
#define	SOH			1
#define	STX			2
#define	ETX			3
#define ETB			0x17
#define	ACK			6
#define NAK			0x15
#define CMD_STATUS		0x70
#define CMD_RESET		0x00
#define CMD_FREQ_SET	0x90
#define CMD_BAUD_SET	0x9A
#define CMD_CHIP_ERASE	0x20
#define CMD_BLOCK_ERASE	0x22
#define CMD_PROGRAM		0x40
#define CMD_VERIFY		0x13
#define CMD_BLK_BLANK	0x32
#define CMD_CHECKSUM	0xB0
#define CMD_SECUR_SET	0xA0



/*-----------------------------------------------------------------------
  Device Parameters
-----------------------------------------------------------------------*/

/* Block table for Hx2 devices */
DWORD Blk_Hx2[] = {
	0x00000, 0x0E000, 0x10000, 0x1E000, 0x20000, 0x2E000, 0x3C000, 0x3E000,
	0x40000, 0x4F000, 0x5E000, 0x6D000, 0x7C000, 0x7D000, 0x7E000, 0x7F000,
	0x80000
};

/* Block table for Ix2 devices */
DWORD Blk_Ix2[] = {
	0x00000, 0x02000, 0x10000, 0x12000, 0x20000
};

/* Block table for Jx2 devices */
DWORD Blk_Jx2[] = {
	0x00000, 0x07000, 0x0E000, 0x15000, 0x1C000, 0x1D000, 0x1E000, 0x1F000,
	0x20000, 0x30000, 0x40000, 0x50000, 0x60000, 0x70000, 0x80000, 0x90000,
	0xA0000
};

/* Block table for 4K block devices */
DWORD Blk_4k[] = {
	0x00000, 0x01000, 0x02000, 0x03000, 0x04000, 0x05000, 0x06000, 0x07000,
	0x08000, 0x09000, 0x0A000, 0x0B000, 0x0C000, 0x0D000, 0x0E000, 0x0F000,
	0x10000, 0x11000, 0x12000, 0x13000, 0x14000, 0x15000, 0x16000, 0x17000,
	0x18000, 0x19000, 0x1A000, 0x1B000, 0x1C000, 0x1D000, 0x1E000, 0x1F000,
	0x20000, 0x21000, 0x22000, 0x23000, 0x24000, 0x25000, 0x26000, 0x27000,
	0x28000, 0x29000, 0x2A000, 0x2B000, 0x2C000, 0x2D000, 0x2E000, 0x2F000,
	0x30000, 0x31000, 0x32000, 0x33000, 0x34000, 0x35000, 0x36000, 0x37000,
	0x38000, 0x39000, 0x3A000, 0x3B000, 0x3C000, 0x3D000, 0x3E000, 0x3F000,
	0x40000, 0x41000, 0x42000, 0x43000, 0x44000, 0x45000, 0x46000, 0x47000,
	0x48000, 0x49000, 0x4A000, 0x4B000, 0x4C000, 0x4D000, 0x4E000, 0x4F000,
	0x50000, 0x51000, 0x52000, 0x53000, 0x54000, 0x55000, 0x56000, 0x57000,
	0x58000, 0x59000, 0x5A000, 0x5B000, 0x5C000, 0x5D000, 0x5E000, 0x5F000,
	0x60000, 0x61000, 0x62000, 0x63000, 0x64000, 0x65000, 0x66000, 0x67000,
	0x68000, 0x69000, 0x6A000, 0x6B000, 0x6C000, 0x6D000, 0x6E000, 0x6F000,
	0x70000, 0x71000, 0x72000, 0x73000, 0x74000, 0x75000, 0x76000, 0x77000,
	0x78000, 0x79000, 0x7A000, 0x7B000, 0x7C000, 0x7D000, 0x7E000, 0x7F000,
	0x80000, 0x81000, 0x82000, 0x83000, 0x84000, 0x85000, 0x86000, 0x87000,
	0x88000, 0x89000, 0x8A000, 0x8B000, 0x8C000, 0x8D000, 0x8E000, 0x8F000,
	0x90000, 0x91000, 0x92000, 0x93000, 0x94000, 0x95000, 0x96000, 0x97000,
	0x98000, 0x99000, 0x9A000, 0x9B000, 0x9C000, 0x9D000, 0x9E000, 0x9F000,
	0xA0000, 0xA1000, 0xA2000, 0xA3000, 0xA4000, 0xA5000, 0xA6000, 0xA7000,
	0xA8000, 0xA9000, 0xAA000, 0xAB000, 0xAC000, 0xAD000, 0xAE000, 0xAF000,
	0xB0000, 0xB1000, 0xB2000, 0xB3000, 0xB4000, 0xB5000, 0xB6000, 0xB7000,
	0xB8000, 0xB9000, 0xBA000, 0xBB000, 0xBC000, 0xBD000, 0xBE000, 0xBF000,
	0xC0000, 0xC1000, 0xC2000, 0xC3000, 0xC4000, 0xC5000, 0xC6000, 0xC7000,
	0xC8000, 0xC9000, 0xCA000, 0xCB000, 0xCC000, 0xCD000, 0xCE000, 0xCF000,
	0xD0000, 0xD1000, 0xD2000, 0xD3000, 0xD4000, 0xD5000, 0xD6000, 0xD7000,
	0xD8000, 0xD9000, 0xDA000, 0xDB000, 0xDC000, 0xDD000, 0xDE000, 0xDF000,
	0xE0000, 0xE1000, 0xE2000, 0xE3000, 0xE4000, 0xE5000, 0xE6000, 0xE7000,
	0xE8000, 0xE9000, 0xEA000, 0xEB000, 0xEC000, 0xED000, 0xEE000, 0xEF000,
	0xF0000, 0xF1000, 0xF2000, 0xF3000, 0xF4000, 0xF5000, 0xF6000, 0xF7000,
	0xF8000, 0xF9000, 0xFA000, 0xFB000, 0xFC000, 0xFD000, 0xFE000, 0xFF000
};

/* Block table for 2K block devices */
DWORD Blk_2k[] = {
	0x00000, 0x00800, 0x01000, 0x01800, 0x02000, 0x02800, 0x03000, 0x03800,
	0x04000, 0x04800, 0x05000, 0x05800, 0x06000, 0x06800, 0x07000, 0x07800,
	0x08000, 0x08800, 0x09000, 0x09800, 0x0A000, 0x0A800, 0x0B000, 0x0B800,
	0x0C000, 0x0C800, 0x0D000, 0x0D800, 0x0E000, 0x0E800, 0x0F000, 0x0F800,
	0x10000, 0x10800, 0x11000, 0x11800, 0x12000, 0x12800, 0x13000, 0x13800,
	0x14000, 0x14800, 0x15000, 0x15800, 0x16000, 0x16800, 0x17000, 0x17800,
	0x18000, 0x18800, 0x19000, 0x19800, 0x1A000, 0x1A800, 0x1B000, 0x1B800,
	0x1C000, 0x1C800, 0x1D000, 0x1D800, 0x1E000, 0x1E800, 0x1F000, 0x1F800,
	0x20000, 0x20800, 0x21000, 0x21800, 0x22000, 0x22800, 0x23000, 0x23800,
	0x24000, 0x24800, 0x25000, 0x25800, 0x26000, 0x26800, 0x27000, 0x27800,
	0x28000, 0x28800, 0x29000, 0x29800, 0x2A000, 0x2A800, 0x2B000, 0x2B800,
	0x2C000, 0x2C800, 0x2D000, 0x2D800, 0x2E000, 0x2E800, 0x2F000, 0x2F800,
	0x30000, 0x30800, 0x31000, 0x31800, 0x32000, 0x32800, 0x33000, 0x33800,
	0x34000, 0x34800, 0x35000, 0x35800, 0x36000, 0x36800, 0x37000, 0x37800,
	0x38000, 0x38800, 0x39000, 0x39800, 0x3A000, 0x3A800, 0x3B000, 0x3B800,
	0x3C000, 0x3C800, 0x3D000, 0x3D800, 0x3E000, 0x3E800, 0x3F000, 0x3F800
};



typedef struct _DEVLST {
	DWORD model;
	DWORD flashsize;
	DWORD *blocktbl;
	DWORD cls;
} DEVLST;

DEVLST devices[] = {
	{ 3700, 0x010000, Blk_Hx2, 0},
	{ 3701, 0x020000, Blk_Hx2, 0},
	{ 3702, 0x010000, Blk_Hx2, 0},
	{ 3703, 0x020000, Blk_Hx2, 0},
	{ 3704, 0x040000, Blk_Hx2, 0},
	{ 3706, 0x020000, Blk_Hx2, 0},
	{ 3707, 0x040000, Blk_Hx2, 0},
	{ 3709, 0x020000, Blk_Hx2, 0},
	{ 3710, 0x040000, Blk_Hx2, 0},
	{ 3711, 0x05E000, Blk_Hx2, 0},
	{ 3712, 0x080000, Blk_Hx2, 0},
	{ 3713, 0x010000, Blk_Ix2, 0},
	{ 3714, 0x020000, Blk_Ix2, 0},
	{ 3715, 0x020000, Blk_Jx2, 0},
	{ 3716, 0x040000, Blk_Jx2, 0},
	{ 3717, 0x060000, Blk_Jx2, 0},
	{ 3718, 0x080000, Blk_Jx2, 0},
	{ 3719, 0x0A0000, Blk_Jx2, 0},
	{ 3720, 0x020000, Blk_Jx2, 0},
	{ 3721, 0x040000, Blk_Jx2, 0},
	{ 3722, 0x060000, Blk_Jx2, 0},
	{ 3723, 0x080000, Blk_Jx2, 0},
	{ 3724, 0x0A0000, Blk_Jx2, 0},

	{ 3726, 0x020000, Blk_2k, 1},
	{ 3728, 0x020000, Blk_2k, 1},
	{ 3729, 0x040000, Blk_2k, 1},
	{ 3731, 0x020000, Blk_2k, 1},
	{ 3732, 0x040000, Blk_2k, 1},
	{ 3733, 0x020000, Blk_2k, 1},
	{ 3734, 0x040000, Blk_2k, 1},
	{ 3735, 0x020000, Blk_2k, 1},
	{ 3736, 0x040000, Blk_2k, 1},

	{ 3739, 0x060000, Blk_4k, 1},
	{ 3740, 0x080000, Blk_4k, 1},
	{ 3741, 0x0C0000, Blk_4k, 1},
	{ 3742, 0x100000, Blk_4k, 1},
	{ 3743, 0x060000, Blk_4k, 1},
	{ 3744, 0x080000, Blk_4k, 1},
	{ 3745, 0x0C0000, Blk_4k, 1},
	{ 3746, 0x100000, Blk_4k, 1},
	{ 3737, 0x020000, Blk_2k, 1},
	{ 3738, 0x040000, Blk_2k, 1},
	{ 3792, 0x060000, Blk_4k, 1},
	{ 3793, 0x080000, Blk_4k, 1},

	{ 3841, 0x0C0000, Blk_4k, 1},
	{ 3842, 0x100000, Blk_4k, 1},
	{ 3797, 0x004000, Blk_2k, 1},
	{ 3801, 0x004000, Blk_2k, 1},
	{ 3805, 0x004000, Blk_2k, 1},
	{ 3798, 0x008000, Blk_2k, 1},
	{ 3802, 0x008000, Blk_2k, 1},
	{ 3806, 0x008000, Blk_2k, 1},
	{ 3799, 0x010000, Blk_2k, 1},
	{ 3803, 0x010000, Blk_2k, 1},
	{ 3807, 0x010000, Blk_2k, 1},
	{ 3800, 0x020000, Blk_2k, 1},
	{ 3804, 0x020000, Blk_2k, 1},
	{ 3808, 0x020000, Blk_2k, 1},
	{ 3838, 0x040000, Blk_2k, 1},
	{ 3839, 0x040000, Blk_2k, 1},
	{ 3840, 0x040000, Blk_2k, 1},

	{ 3333, 0x040000, Blk_4k, 1},
	{ 3334, 0x060000, Blk_4k, 1},
	{ 3335, 0x040000, Blk_4k, 1},
	{ 3336, 0x050000, Blk_4k, 1},
	{ 3340, 0x080000, Blk_4k, 1},
	{ 3341, 0x0A0000, Blk_4k, 1},
	{ 3342, 0x0C0000, Blk_4k, 1},
	{ 3343, 0x100000, Blk_4k, 1},
	{ 3344, 0x060000, Blk_4k, 1},
	{ 3345, 0x080000, Blk_4k, 1},
	{ 3346, 0x0A0000, Blk_4k, 1},
	{ 3347, 0x0C0000, Blk_4k, 1},
	{ 3348, 0x100000, Blk_4k, 1},
	{ 3350, 0x080000, Blk_4k, 1},
	{ 3351, 0x0A0000, Blk_4k, 1},
	{ 3352, 0x0C0000, Blk_4k, 1},
	{ 3353, 0x100000, Blk_4k, 1},
	{ 3354, 0x060000, Blk_4k, 1},
	{ 3355, 0x080000, Blk_4k, 1},
	{ 3356, 0x0A0000, Blk_4k, 1},
	{ 3357, 0x0C0000, Blk_4k, 1},
	{ 3358, 0x100000, Blk_4k, 1},
	{ 3364, 0x060000, Blk_4k, 1},
	{ 3365, 0x080000, Blk_4k, 1},
	{ 3366, 0x0A0000, Blk_4k, 1},
	{ 3367, 0x0C0000, Blk_4k, 1},
	{ 3368, 0x100000, Blk_4k, 1},

	{ 0   , 0       , 0     , 0}
};



/*-----------------------------------------------------------------------
  Global variables (initialized by load_commands())
-----------------------------------------------------------------------*/

DWORD FlashSize, *BlockTbl, DevCls;	/* Last address + 1, Ptr to block table, Device class */
DWORD Bps, iBps;			/* 3:9600, 4:19200, 5:31250, 6:38400, 7:76800, 8:153600 */
DWORD Freq;					/* Input clock frequency [kHz] */
DWORD PortNum = 1;			/* COM port number */

DWORD Compare, Erase, Wait;	/* Compare mode, Force chip erase */
DWORD CodeSize;				/* Loaded program code size */
BYTE CodeBuff[MAX_FLASH];	/* Program code buffer */




DCB Dcb = { sizeof (DCB),
			9600, TRUE, FALSE, FALSE, FALSE,
			DTR_CONTROL_DISABLE, FALSE,
			TRUE, FALSE, FALSE, FALSE, FALSE,
			RTS_CONTROL_DISABLE, FALSE, 0, 0,
			10, 10,
			8, NOPARITY, ONESTOPBIT, '\x11', '\x13', '\xFF', '\xFF', 0 };

COMMTIMEOUTS Commtimeouts1 = { 0, 1, 200, 1, 200};

HANDLE hComm = INVALID_HANDLE_VALUE;



static
const char MessUsage[] =
	"V850ES flash programming tool R0.04b (C)ChaN,2015  http://elm-chan.org/\n\n"
	"Usage: sp850es <switches..> <hex file> [<hex file>] ...\n\n"
    "<hex files> : Load file\n"
	"-erase      : Force chip erase instead of block erase\n"
	"-compare    : Compare without write\n"
	"-dev=<xxxx> : Device number UPD70Fxxxx\n"
	"-freq=<n>   : Oscillator frequecy in kHz (f = 2000-10000)\n"
	"-port=<n>   : COM port number [1] (n = 1-99)\n"
	"-bps=<n>    : Bit rate [9600] (n = 9600,19200,38400,76800,153600)\n"
	"-wait=<n>   : Wait on exit, 1:on error, 2:always\n"
	"-?          : Show this message\n\n"
	"Supprted Device:\n"
	"UPD70F37(00,01,02,03,04,06,07,09,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,26,28,29,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,92,93,97,98,99)\n"
	"UPD70F38(01,02,04,05,06,07,08,38,39,40,41,42)\n"
	"UPD70F33(33,34,35,36,40,41,42,43,44,45,46,47,48,50,51,52,53,54,55,56,57,58,64,65,66,67,68)"
;



/*-----------------------------------------------------------------------
  Hex format manupilations
-----------------------------------------------------------------------*/


/* Pick a hexdecimal value from hex record */

DWORD get_valh (char **lp,	/* pointer to line read pointer */
			   int count, 	/* number of digits to get (2,4,6,8) */
			   BYTE *sum)	/* byte check sum */
{
	DWORD val = 0;
	BYTE n;


	do {
		n = *(*lp)++;
		if ((n -= '0') >= 10) {
			if ((n -= 7) < 10) return(0xFFFFFFFF);
			if (n > 0xF) return(0xFFFFFFFF);
		}
		val = (val << 4) + n;
		if (count & 1) *sum += (BYTE)val;
	} while (--count);
	return(val);
}




/* Load Intel/Motorola hex file into data buffer */ 

long input_hexfile (FILE *fp,			/* input stream */
				   BYTE *buffer,		/* data input buffer */
				   DWORD buffsize,		/* size of data buffer */
				   DWORD *datasize)		/* effective data size in the input buffer */
{
	char line[600];			/* line input buffer */
	char *lp;				/* line read pointer */
	long lnum = 0;			/* input line number */
	WORD seg = 0, hadr = 0;	/* address expantion values for intel hex */
	DWORD addr, count, n;
	BYTE sum;


	while (fgets(line, sizeof line, fp) != NULL) {
		lnum++;
		lp = &line[1]; sum = 0;

		if (line[0] == ':') {	/* Intel Hex format */
			if ((count = get_valh(&lp, 2, &sum)) > 0xFF) return(lnum);	/* byte count */
			if ((addr = get_valh(&lp, 4, &sum)) > 0xFFFF) return(lnum);	/* offset */

			switch (get_valh(&lp, 2, &sum)) {	/* block type? */
				case 0x00 :	/* data block */
					addr += (seg << 4) + (hadr << 16);
					while (count--) {
						if ((n = get_valh(&lp, 2, &sum)) > 0xFF) return(lnum);
						if (addr >= buffsize) continue;	/* clip by buffer size */
						buffer[addr++] = (BYTE)n;		/* store the data */
						if (addr > *datasize)			/* update data size information */
							*datasize = addr;
					}
					break;

				case 0x01 :	/* end block */
					if (count) return(lnum);
					break;

				case 0x02 :	/* segment block */
					if (count != 2) return(lnum);
					if ((seg = (WORD)get_valh(&lp, 4, &sum)) == 0xFFFF) return(lnum);
					break;

				case 0x03 :	/* program start address (segment:offset) */
					if (count != 4) return lnum;
					get_valh(&lp, 8, &sum);
					break;

				case 0x04 :	/* high address block */
					if (count != 2) return(lnum);
					if ((hadr = (WORD)get_valh(&lp, 4, &sum)) == 0xFFFF) return(lnum);
					break;

				case 0x05 :	/* program start address (linear) */
					if (count != 4) return lnum;
					get_valh(&lp, 8, &sum);
					break;

				default:	/* invalid block */
					return(lnum);
			} /* switch */
			if (get_valh(&lp, 2, &sum) > 0xFF) return(lnum);	/* get check sum */
			if (sum) return(lnum);							/* test check sum */
		} /* if */

		if (line[0] == 'S') {	/* Motorola S format */
			if ((*lp >= '1')&&(*lp <= '3')) {

				switch (*lp++) {	/* record type? (S1/S2/S3) */
					case '1' :
						if ((count = get_valh(&lp, 2, &sum) - 3) > 0xFF) return(lnum);
						if ((addr = get_valh(&lp, 4, &sum)) == 0xFFFFFFFF) return(lnum);
						break;
					case '2' :
						if ((count = get_valh(&lp, 2, &sum) - 4) > 0xFF) return(lnum);
						if ((addr = get_valh(&lp, 6, &sum)) == 0xFFFFFFFF) return(lnum);
						break;
					default :
						if ((count = get_valh(&lp, 2, &sum) - 5) > 0xFF) return(lnum);
						if ((addr = get_valh(&lp, 8, &sum)) == 0xFFFFFFFF) return(lnum);
				}
				while (count--) {
					if ((n = get_valh(&lp, 2, &sum)) > 0xFF) return(lnum);
					if (addr >= buffsize) continue;	/* clip by buffer size */
					buffer[addr++] = (BYTE)n;		/* store the data */
					if (addr > *datasize)			/* update data size information */
						*datasize = addr;
				}
				if (get_valh(&lp, 2, &sum) > 0xFF) return(lnum);	/* get check sum */
				if (sum != 0xFF) return(lnum);					/* test check sum */
			} /* switch */
		} /* if */

	} /* while */

	return( feof(fp) ? 0 : -1 );
}





/*-----------------------------------------------------------------------
  Command line analysis
-----------------------------------------------------------------------*/


int load_commands (int argc, char **argv)
{
	char *cp, *sp, *cmdlst[20], cmdbuff[256], filename[256], *dmy;
	int cmd;
	DWORD n;
	long ln;
	FILE *fp;
	DEVLST *dev;
	DWORD BpsTbl[] = { 9600, 19200, 31250, 38400, 76800, 153600 };


	memset(CodeBuff, 0xFF, sizeof CodeBuff);
	cmd = 0; cp = cmdbuff;

	/* Import ini file as command line parameters */
	if ((fp = fopen(INIFILE, "rt")) == NULL) {
		if (SearchPath(NULL, INIFILE, NULL, sizeof filename, filename, &dmy))
			fp = fopen(filename, "rt");
	}
	if (fp != NULL) {
		while (fgets(cp, cmdbuff + sizeof cmdbuff - cp, fp) != NULL) {
			if (cmd >= (sizeof cmdlst / sizeof cmdlst[0] - 1)) break;
			if (*cp <= ' ') break;
			cmdlst[cmd++] = cp;
			for (sp = cp; *sp > ' '; sp++);
			*sp = '\0'; cp = sp + 1;
		}
		fclose(fp);
	}

	/* Get command line parameters */
	while (--argc && (cmd < (sizeof cmdlst / sizeof cmdlst[0] - 1)))
		cmdlst[cmd++] = *++argv;
	cmdlst[cmd] = NULL;

	/* Analyze command line parameters... */
	for (cmd = 0; cmdlst[cmd] != NULL; cmd++) {
		cp = cmdlst[cmd];

		if (strstr(cp, "-port=") == cp) {	/* COM port number */
			PortNum = atoi(cp+6);
			if (PortNum < 1 || PortNum > 99) {
				printf("Invalid com port number.\n");
				return 8;
			}
			continue;
		}
		if (strstr(cp, "-compare") == cp) {	/* Compare */
			Compare = 1;
			continue;
		}
		if (strstr(cp, "-erase") == cp) {	/* Use chip erase instead of block erase */
			Erase = 1;
			continue;
		}
		if (strstr(cp, "-bps=") == cp) {	/* Speed */
			Bps = atoi(cp+5);
			for (iBps = 3; iBps < 9 && BpsTbl[iBps-3] != Bps; iBps++) ;
			if (iBps >= 9) {
				printf("Invalid bit rate.\n");
				return 8;
			}
			continue;
		}
		if (strstr(cp, "-freq=") == cp) {	/* Frequency */
			Freq = atoi(cp+6);
			if (Freq > 10000 || Freq < 2000) {
				printf("Oscillator frequency %d kHz is out of range.\n", Freq);
				return 8;
			}
			continue;
		}
		if (strstr(cp, "-dev=") == cp) {	/* Device number */
			n = atoi(cp+5);
			for (dev = devices; dev->model && dev->model != n; dev++) ;
			if (!dev->model) {
				printf("The specified device UPD70F%04u is not supported.\n", n);
				return 8;
			}
			FlashSize = dev->flashsize;
			BlockTbl = dev->blocktbl;
			DevCls = dev->cls;
			continue;
		}
		if (strstr(cp, "-wait=") == cp) {	/* Wait */
			n = atoi(cp+6);
			Wait = n;
			continue;
		}
		if (strstr(cp, "-?") == cp) {		/* Help */
			return 9;
		}
		if (strstr(cp, "-") == cp) {	/* Invalid command */
			printf("\"%s\" is an invalid command.\n", cp);
			return 8;
		} else {						/* Program code */
			if ((fp = fopen(cp, "rt")) == NULL) {
				printf("Unable to open \"%s\".\n", cp);
				return 7;
			}
			printf("Loading \"%s\"...", cp);
			ln = input_hexfile(fp, CodeBuff, sizeof CodeBuff, &CodeSize);
			fclose(fp);
			if (ln) {
				if (ln < 0)
					printf(" file access failure.\n");
				else
					printf(" hex format error (%ld).\n", ln);
				return 7;
			}
			printf("\n");
			continue;
		}

	}

	if (!FlashSize || !CodeSize) return 9;
	printf("Size of loaded code is 0x%lX.\n", CodeSize);
	if (CodeSize > FlashSize) {
		printf("Code size exceeds the flash size.\n");
		return 7;
	}
	return 0;
}




/*-----------------------------------------------------------------------
  Programming Code
-----------------------------------------------------------------------*/


void send_frame (BYTE header, BYTE footer, const BYTE* buffer, DWORD count)
{
	BYTE n, s[2];
	DWORD wc;


	/* Flush receiving buffer */
	Commtimeouts1.ReadTotalTimeoutConstant = 0;
	SetCommTimeouts(hComm, &Commtimeouts1);
	do
		ReadFile(hComm, s, 2, &wc, NULL);
	while (wc == 2);

	s[0] = header; s[1] = (BYTE)count;			/* STX|ETX and data count */
	WriteFile(hComm, s, 2, &wc, NULL);

	WriteFile(hComm, buffer, count, &wc, NULL);	/* data */

	n = 0 - (BYTE)count;						/* Sum */
	do
		n -= *buffer++;
	while (--count);

	s[0] = n; s[1] = footer;				/* Sum and ETB|ETX */
	WriteFile(hComm, s, 2, &wc, NULL);
}



BOOL rcvr_frame (BYTE *buff, DWORD lbuff, DWORD tmr)
{
	BYTE d[2], sum;
	DWORD rd, len, n;


	Commtimeouts1.ReadTotalTimeoutConstant = tmr;	/* Set timeout value */
	SetCommTimeouts(hComm, &Commtimeouts1);

	ReadFile(hComm, d, 2, &rd, NULL);				/* Get Heder and Length */
	if (rd != 2 || d[0] != STX || !d[1]) return FALSE;
	len = d[1];
	if (len != lbuff) return FALSE;

	ReadFile(hComm, buff, len, &rd, NULL);			/* Get Data */
	if (rd != len) return FALSE;

	ReadFile(hComm, d, 2, &rd, NULL);				/* Get Sum and Footer */
	if (rd != 2 || d[1] != ETX) return FALSE;

	sum = 0 - (BYTE)len;							/* Check sum error */
	for (n = 0; n < len; n++) sum -= buff[n];
	if (sum != d[0]) return FALSE;

	return TRUE;
}




int open_port (void)
{
	char str[20];


	sprintf(str, "\\.\\COM%u", PortNum);
	hComm = CreateFile(str, GENERIC_READ|GENERIC_WRITE, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	if (hComm == INVALID_HANDLE_VALUE) {
		fprintf(stderr, "%s could not be opened.\n", str);
		return 5;
	}
	SetCommState(hComm, &Dcb);
	return 0;
}



int boot (void)
{
	BYTE buf[6];
	int n, stat = 0;


	/* Release device RESET */
	Dcb.fDtrControl = DTR_CONTROL_ENABLE;	/* ER = ON */
	Dcb.fRtsControl = RTS_CONTROL_ENABLE;	/* RS = ON */
	SetCommState(hComm, &Dcb);
	Sleep(250);								/* Delay 250ms */

	/* Bit rate synchronization */
	printf("Sync...");
	buf[0] = 0;
	WriteFile(hComm, buf, 1, &n, NULL);		/* Send a 0x00 */
	Sleep(50);								/* 50ms */
	WriteFile(hComm, buf, 1, &n, NULL);		/* Send a 0x00 */

	/* RESET command */
	for (n = 16; n; n--) {
		Sleep(50);							/* 50ms */
		buf[0] = CMD_RESET;
		send_frame(SOH, ETX, buf, 1);
		if (rcvr_frame(buf, 1, 100) && buf[0] == ACK) break;
	}
	if (!n) goto boot_fail;

	/* Set frequency if specified */
	if (Freq) {
		printf("Freq...");
		buf[0] = CMD_FREQ_SET;
		for (n = 3; Freq > 999; Freq/=10, n++);
		buf[4] = (BYTE)n;
		buf[3] = (BYTE)(Freq % 10); Freq /=10;
		buf[2] = (BYTE)(Freq % 10); Freq /=10;
		buf[1] = (BYTE)Freq;
		send_frame(SOH, ETX, buf, 5);
		if (!rcvr_frame(buf, 1, 100)) { stat = -1; goto boot_fail; }
		if (buf[0] != ACK) { stat = buf[0]; goto boot_fail; }
	}

	/* Set baud rate if specified */
	if (Bps) {
		printf("Baud...");
		buf[0] = CMD_BAUD_SET;
		buf[1] = (BYTE)iBps;
		send_frame(SOH, ETX, buf, 2);
		Sleep(30);
		Dcb.BaudRate = Bps;
		if (!SetCommState(hComm, &Dcb)) { stat = -2; goto boot_fail; }
		Sleep(30);
		buf[0] = CMD_RESET;
		send_frame(SOH, ETX, buf, 1);
		if (!rcvr_frame(buf, 1, 100)) { stat = -1; goto boot_fail; }
		if (buf[0] != ACK) { stat = buf[0]; goto boot_fail; }
	}

	printf("OK.");
	return 0;

boot_fail:
	printf("failed(%d).\n", stat);		
	return 3;
}




int program (void)
{
	BYTE buf[16], blk;
	DWORD addr, blksize, a, c;
	int stat = 0;


	if (Erase) {	/* Use chip erase instead of block erase */
		printf("\nChip erase...");
		buf[0] = CMD_CHIP_ERASE;
		send_frame(SOH, ETX, buf, 1);
		if (!rcvr_frame(buf, 1, 30000)) { stat = -1; goto prog_fail; }
		if (buf[0] != ACK) { stat = buf[0]; goto prog_fail; }
	}

	for (addr = blk = 0; addr < CodeSize; addr += blksize, blk++) {
		/* Get block size */
		blksize = *(BlockTbl + blk + 1) - *(BlockTbl + blk);
		printf("\nBlock %d (%06X-%06X): ", blk, addr, addr + blksize - 1);

		/* Erase block if chip erase was not performed */
		if (!Erase) {
			printf("Erasing..");
			buf[0] = CMD_BLOCK_ERASE;
			switch (DevCls) {
			case 0:
				buf[1] = blk;
				send_frame(SOH, ETX, buf, 2);
				break;
			case 1:
				buf[1] = (BYTE)(addr >> 16);
				buf[2] = (BYTE)(addr >> 8);
				buf[3] = (BYTE)(addr >> 0);
				buf[4] = (BYTE)((addr + blksize - 1) >> 16);
				buf[5] = (BYTE)((addr + blksize - 1) >> 8);
				buf[6] = (BYTE)((addr + blksize - 1) >> 0);
				send_frame(SOH, ETX, buf, 7);
			}
			if (!rcvr_frame(buf, 1, 15000)) { stat = -1; goto prog_fail; }
			if (buf[0] != ACK) { stat = buf[0]; goto prog_fail; }
		}

		/* Skip to program bottom padded 0xFFs */		
		for (c = a = 0; a < blksize; a++) {
			if (CodeBuff[addr + a] != 0xFF)
				c = a + 1;
		}
		if (c == 0) {
			printf(" Skipped");
			continue;	/* Skip programming if all data in the page is 0xFFs */
		}
		c = (c + 255) & 0xFFFF00;

		/* Write block data */
		printf("Writing..");
		buf[0] = CMD_PROGRAM;
		buf[1] = (BYTE)(addr >> 16);
		buf[2] = (BYTE)(addr >> 8);
		buf[3] = (BYTE)(addr);
		buf[4] = (BYTE)((addr + c - 1) >> 16);
		buf[5] = (BYTE)((addr + c - 1) >> 8);
		buf[6] = (BYTE)((addr + c - 1));
		send_frame(SOH, ETX, buf, 7);
		if (!rcvr_frame(buf, 1, 100)) { stat = -1; goto prog_fail; }
		if (buf[0] != ACK) { stat = buf[0]; goto prog_fail; }
		for (a = addr; c; a += 256, c -= 256) {
			send_frame(STX, (BYTE)((c > 256) ? ETB : ETX), &CodeBuff[a], 256);
			if (!rcvr_frame(buf, 2, 300)) { stat = -1; goto prog_fail; }
			if (buf[0] != ACK || buf[1] != ACK) { stat = buf[0]; goto prog_fail; }
		}
		printf(".");
		if (!rcvr_frame(buf, 1, 9000)) { stat = -1; goto prog_fail; }
		if (buf[0] != ACK) { stat = buf[0]; goto prog_fail; }
	}
	printf("\nCompleted.\n");
	return 0;

prog_fail:
	printf("failed(%d).\n", stat);		
	return 3;
}




int compare (void)
{
	BYTE buf[8], blk;
	DWORD addr, blksize, a, c;
	int stat = 0;


	for (addr = blk = 0; addr < CodeSize; addr += blksize, blk++) {
		/* Get block size */
		blksize = *(BlockTbl+blk+1) - *(BlockTbl+blk);
		c = blksize;

		/* Compare block data */
		printf("\nCompareing %06X-%06X...", addr, addr+blksize-1);
		buf[0] = CMD_VERIFY;
		buf[1] = (BYTE)(addr >> 16);
		buf[2] = (BYTE)(addr >> 8);
		buf[3] = (BYTE)(addr);
		buf[4] = (BYTE)((addr+c-1) >> 16);
		buf[5] = (BYTE)((addr+c-1) >> 8);
		buf[6] = (BYTE)((addr+c-1));
		send_frame(SOH, ETX, buf, 7);
		if (!rcvr_frame(buf, 1, 100)) { stat = -1; goto comp_fail; }
		if (buf[0] != ACK) { stat = buf[0]; goto comp_fail; }

		for (a = addr; c; a += 256, c -= 256) {
			send_frame(STX, (BYTE)((c > 256) ? ETB : ETX), &CodeBuff[a], 256);
			if (!rcvr_frame(buf, 2, 100)) { stat = -1; goto comp_fail; }
			if (buf[0] != ACK || buf[1] != ACK) { stat = buf[0]; goto comp_fail; }
		}
	}
	printf("\nNo difference.\n");
	return 0;

comp_fail:
	printf("failed(%d).\n", stat);
	return 3;
}




void pause (
	int rc
)
{
	if ((Wait == 2) || ((Wait == 1) && rc)) {
		printf("\nType Enter to exit...");
		getchar();
	}
}

	
	
int main (int argc, char **argv)
{
	int rc;


	rc = load_commands(argc, argv);
	if (!rc) { 
		rc = open_port();
		if (!rc) {
			rc = boot();
			if (!rc)
				rc = Compare ? compare() : program();
		}
	}
	if (rc == 9) printf(MessUsage);
	pause(rc);

	return rc;
}

