/*----------------------------------------------------------------------------/
/  STM32SP - Flash Programmer for STM32 Microcontrollers      R0.01a
/  http://elm-chan.org/
/-----------------------------------------------------------------------------/
/ STM32SP is a flash programming software for ST Microelectronics STM32 family
/ MCUs. It is a Free Software opened under license policy of GNU GPL.

   Copyright (C) 2013, ChaN, all right reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.

   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.

   * Neither the name of the copyright holders nor the names of
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. 
*/



#include <stdio.h>
#include <string.h>
#include <windows.h>


#define INIFILE "stm32sp.ini"
#define MESS(str)	fputs(str, stderr)
#define FLASHBASE	0x08000000
#define ACK		0x79
#define NAK		0x1F
#define	LD_DWORD(ptr)		(DWORD)(((DWORD)*((BYTE*)(ptr)+3)<<24)|((DWORD)*((BYTE*)(ptr)+2)<<16)|((WORD)*((BYTE*)(ptr)+1)<<8)|*(BYTE*)(ptr))
#define	ST_DWORD(ptr,val)	*(BYTE*)(ptr)=(BYTE)(val); *((BYTE*)(ptr)+1)=(BYTE)((WORD)(val)>>8); *((BYTE*)(ptr)+2)=(BYTE)((DWORD)(val)>>16); *((BYTE*)(ptr)+3)=(BYTE)((DWORD)(val)>>24)


const char *Usage =
	"STM32SP - STM32 Serial Programming tool R0.01a (C)ChaN, 2013\n"
	"\n"
	"Write flash memory:    <hex file> [<hex file>] ...\n"
	"Read flash memory:     -R<size>\n"
	"Port number and speed: -P<n>[:<bps>]\n"
	"Lock device:           -L\n"
	"Unlock device:         -U\n"
	"Signal polarity:       -C<flag> (see stm32sp.ini)\n"
	"Wait on exit:          -W<mode> (see stm32sp.ini)\n"
	;

BYTE CmdList[20], Version[2], ChipID[2];

DWORD AddrRange[2];		/* Loaded address range {lowest, highest} */
BYTE Buffer[0x80000];	/* Flash data buffer (512K) */

UINT Port = 1;			/* -p<port> Port numnber */
UINT Baud = 115200;		/* -p<port>:<bps> Bit rate */
UINT Pause;				/* -w<mode> Pause before exit program */
UINT Pol;				/* -c<flag> Invert signal polarity (b0:ER, b1:RS) */
UINT Lock;				/* -l Lock/Unlock device (read protection) */
UINT Read;				/* -r<size> Read operation */




/*-----------------------------------------------------------------------
  Search and Open configuration file
-----------------------------------------------------------------------*/

static
FILE *open_cfgfile(char *filename)
{
	FILE *fp;
	char filepath[256], *dmy;


	if ((fp = fopen(filename, "rt")) != NULL) return fp;
	if (SearchPath(NULL, filename, NULL, sizeof filepath, filepath, &dmy)) {
		if ((fp = fopen(filepath, "rt")) != NULL) return fp;
	}
	return NULL;
}




/*-----------------------------------------------------------------------
  Hex format manupilations
-----------------------------------------------------------------------*/


/* Pick a hexdecimal value from hex record */

DWORD get_valh (
	char **lp,	/* pointer to line read pointer */
	int count, 	/* number of digits to get (2,4,6,8) */
	BYTE *sum	/* byte check sum */
) {
	DWORD val = 0;
	BYTE n;


	do {
		n = *(*lp)++;
		if ((n -= '0') >= 10) {
			if ((n -= 7) < 10) return 0xFFFFFFFF;
			if (n > 0xF) return 0xFFFFFFFF;
		}
		val = (val << 4) + n;
		if (count & 1) *sum += (BYTE)val;
	} while (--count);
	return val;
}




/* Load Intel Hex and Motorola S format file into data buffer */ 

long input_hexfile (
	FILE* fp,			/* input stream */
	BYTE* buffer,		/* data input buffer */
	DWORD buffsize,		/* size of data buffer */
	DWORD* range		/* effective data size in the input buffer */
) {
	char line[600];			/* line input buffer */
	char *lp;				/* line read pointer */
	long lnum = 0;			/* input line number */
	WORD seg = 0, hadr = 0;	/* address expantion values for intel hex */
	DWORD addr, count, n;
	BYTE sum;


	while (fgets(line, sizeof line, fp) != NULL) {
		lnum++;
		lp = &line[1]; sum = 0;

		if (line[0] == ':') {	/* Intel Hex format */
			if ((count = get_valh(&lp, 2, &sum)) > 0xFF) return lnum;	/* byte count */
			if ((addr = get_valh(&lp, 4, &sum)) > 0xFFFF) return lnum;	/* offset */

			switch (get_valh(&lp, 2, &sum)) {	/* block type? */
				case 0x00 :	/* data */
					addr += (seg << 4) + (hadr << 16);
					addr -= FLASHBASE;
					while (count--) {
						n = get_valh(&lp, 2, &sum);		/* pick a byte */
						if (n > 0xFF) return lnum;
						if (addr >= buffsize) continue;	/* clip by buffer size */
						if (addr > range[1]) range[1] = addr;	/* update data size information */
						if (addr < range[0]) range[0] = addr;
						buffer[addr++] = (BYTE)n;		/* store the data */
					}
					break;

				case 0x01 :	/* end */
					if (count != 0) return lnum;
					break;

				case 0x02 :	/* segment base [19:4] */
					if (count != 2) return lnum;
					seg = (WORD)get_valh(&lp, 4, &sum);
					if (seg == 0xFFFF) return lnum;
					break;

				case 0x03 :	/* program start address (segment:offset) */
					if (count != 4) return lnum;
					get_valh(&lp, 8, &sum);
					break;

				case 0x04 :	/* high address base [31:16] */
					if (count != 2) return lnum;
					hadr = (WORD)get_valh(&lp, 4, &sum);
					if (hadr == 0xFFFF) return lnum;
					break;

				case 0x05 :	/* program start address (linear) */
					if (count != 4) return lnum;
					get_valh(&lp, 8, &sum);
					break;

				default:	/* invalid block */
					return lnum;
			} /* switch */
			if (get_valh(&lp, 2, &sum) > 0xFF) return lnum;	/* get check sum */
			if (sum) return lnum;							/* test check sum */
			continue;
		} /* if */

		if (line[0] == 'S') {	/* Motorola S format */
			if ((*lp >= '1')&&(*lp <= '3')) {

				switch (*lp++) {	/* record type? (S1/S2/S3) */
					case '1' :
						count = get_valh(&lp, 2, &sum) - 3;
						if (count > 0xFF) return lnum;
						addr = get_valh(&lp, 4, &sum);
						if (addr > 0xFFFF) return lnum;
						break;
					case '2' :
						count = get_valh(&lp, 2, &sum) - 4;
						if (count > 0xFF) return lnum;
						addr = get_valh(&lp, 6, &sum);
						if (addr > 0xFFFFFF) return lnum;
						break;
					default :
						count = get_valh(&lp, 2, &sum) - 5;
						if (count > 0xFF) return lnum;
						addr = get_valh(&lp, 8, &sum);
						if (addr == 0xFFFFFFFF) return lnum;
				}
				addr -= FLASHBASE;
				while (count--) {
					n = get_valh(&lp, 2, &sum);
					if (n > 0xFF) return lnum;
					if (addr >= buffsize) continue;	/* clip by buffer size */
					if (addr > range[1]) range[1] = addr;	/* update data size information */
					if (addr < range[0]) range[0] = addr;
					buffer[addr++] = (BYTE)n;		/* store the data */
				}
				if (get_valh(&lp, 2, &sum) > 0xFF) return lnum;	/* get check sum */
				if (sum != 0xFF) return lnum;					/* test check sum */
			} /* switch */
			continue;
		} /* if */

		if (line[0] >= ' ') return lnum;
	} /* while */

	return feof(fp) ? 0 : -1;
}



/* Put an Intel Hex data block */

void put_hexline (
	FILE* fp,			/* output stream */
	const BYTE* buffer,	/* pointer to data buffer */
	WORD ofs,			/* block offset address */
	BYTE count,			/* data byte count */
	BYTE type			/* block type */
) {
	BYTE sum;

	/* Byte count, Offset address and Record type */
	fprintf(fp, ":%02X%04X%02X", count, ofs, type);
	sum = count + (ofs >> 8) + ofs + type;

	/* Data bytes */
	while (count--) {
		fprintf(fp, "%02X", *buffer);
		sum += *buffer++;
	}

	/* Check sum */
	fprintf(fp, "%02X\n", (BYTE)-sum);
}



/* Output data in Intel Hex format */

void output_ihex (
	FILE* fp,			/* output stream */
	const BYTE *buffer,	/* pointer to data buffer */
	DWORD datasize,		/* number of bytes to be output */
	BYTE blocksize		/* HEX block size (1,2,4,..,128) */
) {
	WORD ofs = 0;
	BYTE hadr[2] = {0x08, 0}, d, n;
	DWORD bc = datasize;


	while (bc) {
		if (ofs == 0) {	/* A16 changed? */
			put_hexline(fp, hadr, 0, 2, 4);
			hadr[1]++;
		}
		if (bc >= blocksize) {	/* full data block */
			for (d = 0xFF, n = 0; n < blocksize; n++) d &= *(buffer+n);
			if (d != 0xFF) put_hexline(fp, buffer, ofs, blocksize, 0);
			buffer += blocksize;
			bc -= blocksize;
			ofs += blocksize;
		} else {				/* fractional data block */
			for (d = 0xFF, n = 0; n < bc; n++) d &= *(buffer+n);
			if (d != 0xFF) put_hexline(fp, buffer, ofs, (BYTE)bc, 0);
			bc = 0;
		}
	}

	put_hexline(fp, NULL, 0, 0, 1);	/* End block */
}



/*-----------------------------------------------------------------------
  Command line analysis
-----------------------------------------------------------------------*/


static
int load_commands (int argc, char** argv)
{
	char *cp, *cmdlst[10], cmdbuff[256];
	int cmd;
	FILE *fp;
	long n;


	/* Clear flash data buffer */
	memset(Buffer, 0xFF, sizeof Buffer);		/* Default value (if not loaded) */
	AddrRange[0] = sizeof Buffer;	/* Lowest address */
	AddrRange[1] = 0;				/* Highest address */

	cmd = 0; cp = cmdbuff;

	/* Import ini file as command line parameters */
	fp = open_cfgfile(INIFILE);
	if (fp) {
		while (fgets(cp, cmdbuff + sizeof cmdbuff - cp, fp) != NULL) {
			if (cmd >= (sizeof cmdlst / sizeof cmdlst[0] - 1)) break;
			if (*cp <= ' ') break;
			while (cp[strlen(cp) - 1] <= ' ') cp[strlen(cp) - 1] = 0;
			cmdlst[cmd++] = cp; cp += strlen(cp) + 1;
		}
		fclose(fp);
	}

	/* Get command line parameters */
	while (--argc && (cmd < (sizeof cmdlst / sizeof cmdlst[0] - 1)))
		cmdlst[cmd++] = *++argv;
	cmdlst[cmd] = NULL;

	/* Analyze command line parameters... */
	for (cmd = 0; cmdlst[cmd] != NULL; cmd++) {
		cp = cmdlst[cmd];

		if (*cp == '-') {	/* Command switches... */
			cp++;
			switch (tolower(*cp++)) {
				case 'p' :	/* -p<num>[:<bps>] (control port and bit rate) */
					Port = strtoul(cp, &cp, 10);
					if(*cp == ':')
						Baud = strtoul(cp+1, &cp, 10);
					break;

				case 'r' :	/* -r<size> (read command) */
					Read = strtoul(cp, &cp, 10);
					break;

				case 'w' :	/* -w<num> (pause before exit 1:on error or 2:always) */
					Pause = strtoul(cp, &cp, 10);
					break;

				case 'c' :	/* -c<flag> (polarity of control signals) */
					Pol = strtoul(cp, &cp, 10);
					break;

				case 'l' :	/* -l Lock device */
					Lock = 1;
					break;

				case 'u' :	/* -u Unlock device */
					Lock = 2;
					break;

				default :	/* invalid command */
					return 1;
			} /* switch */
			if(*cp >= ' ') return 1;	/* option trails garbage */
		} /* if */

		else {	/* HEX Files */
			fprintf(stderr, "Loading \"%s\"...", cp);
			if ((fp = fopen(cp, "rt")) == NULL) {
				fprintf(stderr, "Unable to open.\n");
				return 2;
			}
			n = input_hexfile(fp, Buffer, sizeof Buffer, AddrRange);
			fclose(fp);
			if (n) {
				if (n < 0) {
					fprintf(stderr, "file access failure.\n");
				} else {
					fprintf(stderr, "hex format error at line %ld.\n", n);
				}
				return 2;
			}
			fprintf(stderr, "passed.\n");
		} /* else */

	} /* for */

	return 0;
}



/*-----------------------------------------------------------------------
  STM32 programming controls
-----------------------------------------------------------------------*/

static
int rcv_ack (
	HANDLE h
)
{
	BYTE d;
	DWORD nr;

	ReadFile(h, &d, 1, &nr, NULL);
	return (nr == 1 && d == ACK);
}



static
int enter_ispmode (
	HANDLE* com
)
{
	DCB dcb = { sizeof (DCB),
				Baud, TRUE, FALSE, FALSE, FALSE,
				DTR_CONTROL_DISABLE, FALSE,
				TRUE, FALSE, FALSE, FALSE, FALSE,
				RTS_CONTROL_DISABLE, FALSE, 0, 0,
				10, 10,
				8, EVENPARITY, ONESTOPBIT, '\x11', '\x13', '\xFF', '\xFF', 0 };
	COMMTIMEOUTS ct1 = { 0, 1, 200, 1, 200};
	char str[20];
	DWORD nw, nr, n;


	/* Open communication port */
	sprintf(str, "\\\\.\\COM%u", Port);
	*com = CreateFile(str, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (*com == INVALID_HANDLE_VALUE) {
		fprintf(stderr, "%s could not be opened.\n", str);
		return 5;
	}
	SetCommState(*com, &dcb);		/* Set serial bit rate */
	SetCommTimeouts(*com, &ct1);	/* Set processing timeout 200m sec */
	EscapeCommFunction(*com, (Pol & 2) ? SETRTS : CLRRTS);	/* Set BOOT pin high if RTS controls it */

	MESS("Entering boot mode...");

	for (n = 3; n; n--) {
		/* Reset the device if DTR signal controls RESET pin */
		EscapeCommFunction(*com, (Pol & 1) ? SETDTR : CLRDTR);
		Sleep(50);
		EscapeCommFunction(*com, (Pol & 1) ? CLRDTR : SETDTR);
		Sleep(100);
		PurgeComm(*com, PURGE_RXABORT | PURGE_RXCLEAR);
		WriteFile(*com, "\x7F", 1, &nw, NULL);
		if (rcv_ack(*com)) break;
	}

	if (n) {
		WriteFile(*com, "\x00\xFF", 2, &nw, NULL);
		if (rcv_ack(*com)) {
			ReadFile(*com, CmdList, 2, &nr, NULL);
			if (nr == 2 && CmdList[0] <= sizeof CmdList - 2) {
				ReadFile(*com, &CmdList[2], CmdList[0], &nr, NULL);
				if (nr == CmdList[0] && rcv_ack(*com)) {
					MESS("\n");
					return 0;
				}
			}
		}
	}

	MESS("failed to start the USART boot loader.\n");
	return 6;
}



static
void exit_ispmode (
	HANDLE com
)
{
	/* Apply a reset to the device if DTR/RTS controls RESET/BOOT pin */
	EscapeCommFunction(com, (Pol & 2) ? CLRRTS : SETRTS);	/* Set BOOT pin low */
	EscapeCommFunction(com, (Pol & 1) ? SETDTR : CLRDTR);	/* Set RESET pin low */
	Sleep(50);
	EscapeCommFunction(com, (Pol & 1) ? CLRDTR : SETDTR);	/* Set RESET pin high */
}




static
int read_flash (
	HANDLE com,
	BYTE* buffer,
	UINT sz_read
)
{
	DWORD ba, ra, nw, nr, n;
	BYTE buf[8], sum;


	MESS("Reading");
	ba = 0;
	do {
		if ((ba % 4096) == 0) MESS(".");
		WriteFile(com, "\x11\xEE", 2, &nw, NULL);
		if (!rcv_ack(com)) break;
		ra = ba + FLASHBASE;
		for (n = sum = 0; n < 4; n++) {
			buf[n] = (BYTE)(ra >> ((3 - n) * 8));
			sum ^= buf[n];
		}
		buf[n] = sum;
		WriteFile(com, buf, 5, &nw, NULL);
		if (!rcv_ack(com)) break;
		WriteFile(com, "\xFF\x00", 2, &nw, NULL);
		if (!rcv_ack(com)) break;
		ReadFile(com, &Buffer[ba], 256, &nr, NULL);
		if (nr != 256) break;
		ba += 256;
	} while (ba < sz_read);

	if (ba < sz_read) {
		MESS("failed.\n");
		return 3;
	} else {
		MESS("passed.\n");
		return 0;
	}
}




static
int erase_flash (
	HANDLE com
)
{
	COMMTIMEOUTS ct1 = { 0, 1, 2000, 1, 250},
				 ct2 = { 0, 1, 200, 1, 200};
	int rc = 0;
	BYTE i, cmd[2];
	DWORD nw;


	for (i = 0; i < CmdList[0]; i++) {
		if (CmdList[2+i] == 0x43) {
			cmd[0] = 0x43; break;
		}
		if (CmdList[2+i] == 0x44) {
			cmd[0] = 0x44; break;
		}
	}
	if (i >= CmdList[0]) {
		MESS("Could not erase flash memory.");
		return 4;
	}
	
	MESS("Erasing...");
	
	cmd[1] = ~cmd[0]; 
	WriteFile(com, cmd, 2, &nw, NULL);
	if (!rcv_ack(com)) rc = 4;
	if (!rc) {
		SetCommTimeouts(com, &ct1);	/* Set processing timeout 2000 msec */
		if (cmd[0] == 0x43) {
			WriteFile(com, "\xFF\x00", 2, &nw, NULL);
		} else {
			WriteFile(com, "\xFF\xFF\x00", 3, &nw, NULL);
		}
		if (!rcv_ack(com)) rc = 4;
		SetCommTimeouts(com, &ct2);	/* Set processing timeout 200 msec */
	}
	MESS(rc ? "failed.\n" : "passed.\n");
	return rc;
}




static
int lock_device (
	HANDLE com,
	UINT lock
)
{
	COMMTIMEOUTS ct1 = { 0, 1, 2000, 1, 250},
				 ct2 = { 0, 1, 200, 1, 200};
	int rc = 0;
	DWORD nw;


	SetCommTimeouts(com, &ct1);	/* Set processing timeout 2000 msec */
	switch (Lock) {
	case 1:
		MESS("Lock device...");
		WriteFile(com, "\x82\x7D", 2, &nw, NULL);
		break;
	case 2:
		MESS("Unlock device...");
		WriteFile(com, "\x92\x6D", 2, &nw, NULL);
		break;
	}
	if (!rcv_ack(com) || !rcv_ack(com)) rc = 4;
	SetCommTimeouts(com, &ct2);	/* Set processing timeout 200 msec */
	MESS(rc ? "failed.\n" : "passed.\n");
	return rc;
}




static
int write_flash (
	HANDLE com,
	const BYTE* buffer,
	DWORD range[]
)
{
	DWORD ba, ra, nw, nr, n;
	BYTE buf[8], sum;


	MESS("Writing");
	ba = 0;
	do {
		if ((ba % 4096) == 0) MESS(".");
		WriteFile(com, "\x31\xCE", 2, &nw, NULL);
		if (!rcv_ack(com)) break;
		ra = ba + FLASHBASE;
		for (n = sum = 0; n < 4; n++) {
			buf[n] = (BYTE)(ra >> ((3 - n) * 8));
			sum ^= buf[n];
		}
		buf[n] = sum;
		WriteFile(com, buf, 5, &nw, NULL);
		if (!rcv_ack(com)) break;
		sum = 0xFF;
		WriteFile(com, &sum, 1, &nw, NULL);
		WriteFile(com, &Buffer[ba], 256, &nr, NULL);
		for (n = 0; n < 256; n++) sum ^= Buffer[ba + n];
		WriteFile(com, &sum, 1, &nw, NULL);
		if (!rcv_ack(com)) break;
		ba += 256;
	} while (ba < range[1]);

	if (ba < range[1]) {
		MESS("failed.\n");
		return 3;
	} else {
		MESS("passed.\n");
		return 0;
	}
}



void pause (
	int rc
)
{
	if ((Pause == 2) || ((Pause == 1) && rc)) {
		MESS("\nType Enter to exit...");
		getchar();
	}
}



int main (int argc, char** argv)
{
	int rc;
	HANDLE hcom;


	rc = load_commands(argc, argv);
	if (rc) {
		if (rc == 1) MESS(Usage);
		pause(rc);
		return rc;
	}

	if (Lock) {
		rc = enter_ispmode(&hcom);
		if (!rc) {
			rc = lock_device(hcom, Lock);
			exit_ispmode(hcom);
		}
		pause(rc);
		return rc;
	}

	if (Read) {
		Read *= 1024;
		if (Read > sizeof Buffer) Read = sizeof Buffer;
		rc = enter_ispmode(&hcom);
		if (!rc) {
			rc = read_flash(hcom, Buffer, Read);
			if (!rc) {
				output_ihex(stdout, Buffer, Read, 32);
			}
			exit_ispmode(hcom);
		}
		pause(rc);
		return rc;
	}

	if (AddrRange[1] == 0) {	/* Check if any data is loaded */
		MESS(Usage);
		rc = 1;
	}
	if (!rc) {
		fprintf(stderr, "Loaded address range is %05X-%05X.\n", AddrRange[0] + FLASHBASE, AddrRange[1] + FLASHBASE);
		if (AddrRange[0] != 0) {	/* Check if vector area is loaded */
			MESS("Vector table is not loaded.\n");
			rc = 1;
		}
	}
	if (!rc)
		rc = enter_ispmode(&hcom);	/* Open port, enter ispmode and detect device type */
	if (!rc) {
		rc = erase_flash(hcom);
		if (!rc)
			rc = write_flash(hcom, Buffer, AddrRange);
		exit_ispmode(hcom);
	}
	pause(rc);
	return rc;
}

